{% extends "!footer.html" %}
{%- block contentinfo %}
{{ super }}
This documentation is licensed under the highly permissive
<a href="https://github.com/befelix/pydantic_sweep/blob/main/docs/LICENSE">
BSD Zero Clause License</a>, enabling you to copy & paste code without attribution.
<br><br>
{% endblock %}